"""Unittests for SaveState."""

import os
import shutil
import pickle
import struct
import unittest
import savestate
from argparse import Namespace


class TestSaveState(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.tempdir = "testdir"
        cls.tempfile = "testfile"
        cls.filepath = os.path.join(cls.tempdir, cls.tempfile)

    @classmethod
    def tearDownClass(cls) -> None:
        if os.path.exists(cls.tempdir):  # noqa
            shutil.rmtree(cls.tempdir)  # noqa

    def setUp(self) -> None:
        if os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)

        os.mkdir(self.tempdir)
        self.db = savestate.open(filename=self.filepath, flag="n")

    def tearDown(self) -> None:
        if self.db.is_open:
            self.db.close()

    def truncate_data_file(self, bytes_from_end: int) -> None:
        with open(self.db.filepath, "rb") as f:
            data = f.read()
        with open(self.db.filepath, "wb") as f:
            f.write(data[:-bytes_from_end])

    # --- Tests -------------------------------------------------------------

    def test_file_identifier_gets_added(self):
        self.assertEqual(self.db.filename[-10:], ".savestate")

    def test_flags(self):
        self.db.close()
        shutil.rmtree(self.tempdir)
        os.mkdir(self.tempdir)

        # Improper flag
        self.assertRaises(ValueError, savestate.open, filename=self.filepath, flag="foo")

        # Test that file can't be created on these modes.
        self.assertRaises(savestate.SaveStateError, savestate.open, filename=self.filepath, flag="r")
        self.assertRaises(savestate.SaveStateError, savestate.open, filename=self.filepath, flag="w")

        try:
            self.db = savestate.open(filename=self.filepath, flag="c")
        except savestate.SaveStateError:
            self.fail("Opening SaveState with flag 'c' when file does not exist should not raise a 'SaveStateError'!")

        self.db["foo"] = "bar"
        self.assertEqual("bar", self.db["foo"])
        self.db.close()

        # Test read only mode
        self.db = savestate.open(filename=self.filepath, flag="r")
        # All methods that read-only does should not have
        self.assertFalse(hasattr(self.db, "__setitem__"))
        self.assertFalse(hasattr(self.db, "__delitem__"))
        self.assertFalse(hasattr(self.db, "sync"))
        self.assertFalse(hasattr(self.db, "compact"))
        self.assertFalse(hasattr(self.db, "clear"))
        self.assertFalse(hasattr(self.db, "setdefault"))
        self.assertFalse(hasattr(self.db, "pop"))
        self.assertFalse(hasattr(self.db, "popitem"))
        self.assertFalse(hasattr(self.db, "copy"))
        self.assertFalse(hasattr(self.db, "update"))
        self.assertFalse(hasattr(self.db, "_rename"))
        self.assertFalse(hasattr(self.db, "_write_headers"))
        # Reading is fine
        self.assertEqual("bar", self.db["foo"])
        self.db.close()

        # Test read-write mode
        self.db = savestate.open(filename=self.filepath, flag="w")
        self.db["foo"] = "baz"
        self.assertEqual("baz", self.db["foo"])
        self.db.close()

        # Test create new always mode
        self.db = savestate.open(filename=self.filepath, flag="n")
        self.assertTrue("foo" not in self.db)
        self.db["foo"] = "bar"
        self.assertEqual("bar", self.db["foo"])
        self.db.close()

        shutil.rmtree(self.tempdir)
        os.mkdir(self.tempdir)

        try:
            self.db = savestate.open(filename=self.filepath, flag="n")
        except savestate.SaveStateError:
            self.fail("Opening SaveState with flag 'n' when file does not exist should not raise a 'SaveStateError'!")

    # --- Test setting and getting different types ---------------------------

    def test_bytes(self):
        self.db[b"foo"] = b"bar"
        self.assertEqual(b"bar", self.db[b"foo"])
        self.assertNotEqual("bar", self.db[b"foo"])
        del self.db[b"foo"]
        self.assertTrue(b"foo" not in self.db)

    def test_str(self):
        self.db["foo"] = "bar"
        self.assertEqual("bar", self.db["foo"])
        self.assertNotEqual(b"bar", self.db["foo"])
        del self.db["foo"]
        self.assertTrue("foo" not in self.db)

    def test_int(self):
        self.db[1] = 2
        self.assertEqual(2, self.db[1])
        del self.db[1]
        self.assertTrue(1 not in self.db)

    def test_float(self):
        self.db[0.1] = 0.2
        self.assertEqual(0.2, self.db[0.1])
        del self.db[0.1]
        self.assertTrue(0.1 not in self.db)

    def test_object(self):
        self.db[Namespace(a=b"gsdfff", b=None, c=True)] = Namespace(x=1, y=0x001, z="None")
        self.assertEqual(Namespace(x=1, y=0x001, z="None"), self.db[Namespace(a=b"gsdfff", b=None, c=True)])
        del self.db[Namespace(a=b"gsdfff", b=None, c=True)]
        self.assertTrue(Namespace(a=b"gsdfff", b=None, c=True) not in self.db)

    # --- Test usecase scenarios ---------------------------------------------

    def test_close_and_reopen(self):
        self.db["foo"] = "bar"
        self.db.close()
        self.db = savestate.open(filename=self.filepath, flag="r")
        self.assertEqual("bar", self.db["foo"])

    def test_get_set_multiple(self):
        self.db["one"] = 1
        self.assertEqual(1, self.db["one"])
        self.db["two"] = 2
        self.assertEqual(2, self.db["two"])
        self.db["three"] = 3
        self.assertEqual(1, self.db["one"])
        self.assertEqual(2, self.db["two"])
        self.assertEqual(3, self.db["three"])

    def test_keyerror_raised_when_key_does_not_exist(self):
        self.assertRaises(KeyError, self.db.__getitem__, "foo")
        self.assertRaises(KeyError, self.db.__delitem__, "foo")
        self.assertRaises(KeyError, self.db.pop, "foo")
        self.assertRaises(KeyError, self.db.popitem)

    def test_updates(self):
        self.db["one"] = "foo"
        self.db["one"] = "bar"
        self.assertEqual("bar", self.db["one"])
        self.db["one"] = "baz"
        self.assertEqual("baz", self.db["one"])

    def test_updates_persist(self):
        self.db["one"] = "foo"
        self.db["one"] = "bar"
        self.db["one"] = "baz"
        self.db.close()
        self.db = savestate.open(filename=self.filepath, flag="r")
        self.assertEqual(self.db["one"], "baz")

    def test_deletes(self):
        self.db["foo"] = "bar"
        del self.db["foo"]
        self.assertTrue("foo" not in self.db)

    def test_deleted_key_not_there_when_reopened(self):
        self.db["one"] = 1
        self.db["two"] = 2
        del self.db["two"]
        self.db.close()

        self.db = savestate.open(filename=self.filepath, flag="r")
        self.assertEqual(self.db["one"], 1)
        self.assertTrue("two" not in self.db)

    def test_multiple_deletes(self):
        self.db["foo"] = "foo"
        del self.db["foo"]
        self.db["foo"] = "foo"
        del self.db["foo"]
        self.db["foo"] = "foo"
        del self.db["foo"]
        self.db["bar"] = "bar"
        self.db.close()

        self.db = savestate.open(filename=self.filepath, flag="r")
        self.assertTrue("foo" not in self.db)
        self.assertEqual(self.db["bar"], "bar")

    def test_update_after_delete(self):
        self.db["one"] = 1
        del self.db["one"]
        self.db["two"] = 2
        self.db["one"] = 3

        self.assertEqual(self.db["two"], 2)
        self.assertEqual(self.db["one"], 3)

    # --- Test savestate methods --------------------------------------------------

    def test_get_method(self):
        self.db["foo"] = "bar"
        self.assertEqual(self.db["foo"], self.db.get("foo"))

    def test_keys_method(self):
        self.db["one"] = 1
        self.db["two"] = 2
        self.db["three"] = 3
        self.assertEqual(self.db.keys(), ["one", "two", "three"])

    def test_values_method(self):
        self.db["one"] = 1
        self.db["two"] = 2
        self.db["three"] = 3
        self.assertEqual(self.db.values(), [1, 2, 3])

    def test_items_method(self):
        self.db["one"] = 1
        self.db["two"] = 2
        self.db["three"] = 3
        self.assertEqual(self.db.items(), [("one", 1), ("two", 2), ("three", 3)])

    def test_close_method(self):
        self.db["foo"] = "bar"
        self.db.close()
        self.assertFalse(hasattr(self.db, "_data_file_descriptor"))
        self.assertRaises(AttributeError, self.db.__getitem__, "foo")

    def test_sync_method(self):
        self.db["foo"] = "bar"
        self.db.sync()
        self.db.close()
        self.assertRaises(AttributeError, self.db.sync)

    def test_contains_method(self):
        self.db["foo"] = "bar"
        self.assertTrue("foo" in self.db)

    def test_iter_method(self):
        self.db["one"] = 1
        self.db["two"] = 2
        self.db["three"] = 3
        self.assertEqual(list(iter(self.db)), ["one", "two", "three"])

    def test_reversed_method(self):
        self.db["one"] = 1
        self.db["two"] = 2
        self.db["three"] = 3
        self.assertEqual(list(reversed(self.db)), ["three", "two", "one"])

    def test_len_method(self):
        self.db["one"] = "foo"
        self.db["one"] = "bar"
        self.db["two"] = "baz"
        self.assertEqual(len(self.db), 2)

    def test_del_method(self):
        try:
            del self.db
            self.db = savestate.open(filename=self.filepath, flag="r")
        except:  # noqa
            self.fail("Database did not close gracefully when deleted")

        try:
            self.db = savestate.open(filename=os.path.join(self.tempdir, "testcopy"), flag="n")
            self.db = savestate.open(filename=self.filepath, flag="r")
        except:  # noqa
            self.fail("Database did not close gracefully when garbage collected")

    def test_context_manager(self):
        self.db.close()

        with savestate.open(filename=self.filepath, flag="n") as db:
            db["one"] = 1
            db["two"] = 2
            db["three"] = 3

        self.db = savestate.open(filename=self.filepath, flag="r")
        self.assertEqual(1, self.db["one"])
        self.assertEqual(2, self.db["two"])
        self.assertEqual(3, self.db["three"])

    def test_pop_method(self):
        self.db["foo"] = "bar"
        value = self.db.pop("foo")
        self.assertEqual(value, "bar")
        self.assertTrue("one" not in self.db)

    def test_popitem_method(self):
        self.db["one"] = 1
        self.db["two"] = 2
        key, value = self.db.popitem()
        self.assertEqual(1, self.db["one"])
        self.assertEqual(key, "two")
        self.assertEqual(value, 2)
        self.assertTrue("two" not in self.db)

    def test_clear_method(self):
        self.db["one"] = "bar"
        self.db["two"] = "bar"
        self.db["three"] = "bar"
        self.db.close()

        before = os.path.getsize(self.db.filepath)

        self.db = savestate.open(filename=self.filepath, flag="c")
        self.db.clear()
        self.assertRaises(KeyError, self.db.__getitem__, "one")
        self.assertRaises(KeyError, self.db.__getitem__, "two")
        self.assertRaises(KeyError, self.db.__getitem__, "three")

        # Clearing should compact the database
        self.db.close()
        after = os.path.getsize(self.db.filepath)
        self.assertLess(after, before)

    def test_setdefault_method(self):
        self.db["foo"] = "bar"
        value = self.db.setdefault("foo", "baz")
        self.assertEqual(value, "bar")
        try:
            value = self.db.setdefault("one", 1)
            self.assertEqual(value, 1)
        except KeyError:
            self.fail("Setdefault should not raise KeyError.")

    def test_update_method(self):
        self.db.update({"foo": "bar", "one": 1}, two=2, foo="baz")
        self.assertEqual(self.db["foo"], "baz")
        self.assertEqual(self.db["one"], 1)
        self.assertEqual(self.db["two"], 2)

    def test_copy_method(self):

        self.db.close()
        copyfile = os.path.join(self.tempdir, "testcopy")

        for mode in ("n", "w", "c"):
            self.db = savestate.open(filename=self.filepath, flag=mode)  # noqa
            self.db["foo"] = "bar"
            new_db = self.db.copy(copyfile)
            self.assertEqual(self.db["foo"], new_db["foo"])
            self.assertEqual(type(self.db), type(new_db))
            new_db.close()
            self.db.close()
            os.remove(new_db.filepath)

        self.db = savestate.open(filename=self.filepath, flag="r")
        self.assertFalse(hasattr(self.db, "copy"))

    def test_covert_to_bytes(self):
        self.assertEqual(self.db._convert_to_bytes("foo"), pickle.dumps("foo", protocol=5))

    def test_covert_from_bytes(self):
        value = pickle.dumps("foo", protocol=5)
        self.assertEqual(self.db._convert_from_bytes(value), pickle.loads(value))

    def test_verify_checksums(self):
        self.db["foo"] = "bar"
        self.db.close()

        self.db = savestate.open(self.filepath, flag="r", verify_checksums=True)

        try:
            _ = self.db["foo"]
        except savestate.SaveStateChecksumError:
            self.fail("Checksum failed.")

    # --- Test compaction ----------------------------------------------------

    def test_compact(self):
        for i in range(10):
            self.db[i] = i
        for i in range(5):
            del self.db[i]

        self.db.close()
        before = os.path.getsize(self.db.filepath)

        self.db = savestate.open(filename=self.filepath, flag="c")
        self.db.compact()
        self.db.close()

        after = os.path.getsize(self.db.filepath)
        self.assertLess(after, before)

    def test_compact_does_not_leave_behind_files(self):
        before = len(os.listdir(self.tempdir))
        for i in range(10):
            self.db[i] = i
        for i in range(10):
            del self.db[i]

        self.db.close()
        self.db = savestate.open(filename=self.filepath, flag="c")
        self.db.compact()

        after = len(os.listdir(self.tempdir))
        self.assertEqual(before, after)

    def test_compact_on_close(self):
        self.db["foo"] = "bar"
        del self.db["foo"]
        self.db.close(compact=True)
        self.db = savestate.open(filename=self.filepath, flag="r")
        # Only header in the file.
        self.assertEqual(self.db._current_offset, 13)

    def test_compact_set_and_get(self):
        for i in range(10):
            self.db[i] = i
        for i in range(5):
            del self.db[i]
        for i in range(5, 10):
            self.db[i] = "foo"

        self.db.compact()

        for i in range(5, 10):
            self.assertEqual(self.db[i], "foo")

        self.assertEqual(len(self.db), 5)

        for i in range(5):
            self.db[i] = i

        self.assertEqual(len(self.db), 10)

    # --- Test corrupted data ------------------------------------------------

    def test_bad_file_identifier_raises_error(self):
        self.db["foo"] = "bar"
        self.db.close()

        with open(self.db.filepath, "r+b") as f:
            data = f.read()

        self.assertEqual(data[:9], b"savestate")

        with open(self.db.filepath, "r+b") as f:
            f.seek(0)
            f.write(b"fdjkasdtj")

        self.assertRaises(savestate.SaveStateLoadError, savestate.open, filename=self.filepath, flag="c")

    def test_incompatible_version_number_raises_error(self):
        self.db["foo"] = "bar"
        self.db.close()

        with open(self.db.filepath, "r+b") as f:
            f.seek(4)
            f.write(struct.pack("!H", 9))

        self.assertRaises(savestate.SaveStateLoadError, savestate.open, filename=self.filepath, flag="c")

    def test_incompatible_picking_version_raises_error(self):
        self.db["foo"] = "bar"
        self.db.close()

        with open(self.db.filepath, "r+b") as f:
            f.seek(6)
            f.write(struct.pack("!H", 1))

        self.assertRaises(savestate.SaveStateLoadError, savestate.open, filename=self.filepath, flag="c")

    def test_warns_but_recovers_from_bad_key_value_size_indicator(self):
        self.db["one"] = "bar"
        self.db["two"] = "bar"
        self.db["three"] = "bar"
        self.db.close()

        before = os.path.getsize(self.db.filepath)

        # Write garbage as the key and value indicator for the second key.
        # The first key should still be read and the db formed from that.
        with open(self.db.filepath, "r+b") as f:
            f.seek(13 + 8 + 18 + 18 + 4)
            f.write(b"\x00" * 8)

        with self.assertWarns(BytesWarning):
            self.db = savestate.open(filename=self.filepath, flag="c")

        self.assertEqual(self.db["one"], "bar")
        self.assertRaises(KeyError, self.db.__getitem__, "two")
        self.assertRaises(KeyError, self.db.__getitem__, "three")

        # Test that compaction removes the corrupted data.
        self.db.close(compact=True)
        after = os.path.getsize(self.db.filepath)
        self.assertLess(after, before)

        self.db = savestate.open(filename=self.filepath, flag="c")
        self.assertEqual(self.db["one"], "bar")

    def test_warns_but_recovers_from_bad_key_data(self):
        self.db["one"] = "bar"
        self.db["two"] = "bar"
        self.db["three"] = "bar"
        self.db.close()

        before = os.path.getsize(self.db.filepath)

        # Write garbage to the beginning of the second key data.
        # The first and third key should still be read and the db formed from that.
        with open(self.db.filepath, "r+b") as f:
            f.seek(13 + 8 + 18 + 18 + 4 + 8)
            f.write(b"\x00" * 8)

        with self.assertWarns(BytesWarning):
            self.db = savestate.open(filename=self.filepath, flag="c")

        self.assertEqual(self.db["one"], "bar")
        self.assertRaises(KeyError, self.db.__getitem__, "two")
        self.assertEqual(self.db["three"], "bar")

        # Test that compaction removes the corrupted data.
        self.db.close(compact=True)
        after = os.path.getsize(self.db.filepath)
        self.assertLess(after, before)

        self.db = savestate.open(filename=self.filepath, flag="c")
        self.assertEqual(self.db["one"], "bar")
        self.assertEqual(self.db["three"], "bar")

    def test_warns_but_recovers_from_bad_value_data(self):
        self.db["one"] = "bar"
        self.db["two"] = "bar"
        self.db["three"] = "bar"
        self.db.close()

        before = os.path.getsize(self.db.filepath)

        # # Write garbage to the beginning of the second value data.
        # The first and third key should still be read and the db formed from that.
        with open(self.db.filepath, "r+b") as f:
            f.seek(13 + 8 + 18 + 18 + 4 + 8 + 18)
            f.write(b"\x00" * 8)

        with self.assertWarns(BytesWarning):
            self.db = savestate.open(filename=self.filepath, flag="c")

        self.assertEqual(self.db["one"], "bar")
        self.assertRaises(KeyError, self.db.__getitem__, "two")
        self.assertEqual(self.db["three"], "bar")

        # Test that compaction removes the corrupted data.
        self.db.close(compact=True)
        after = os.path.getsize(self.db.filepath)
        self.assertLess(after, before)

        self.db = savestate.open(filename=self.filepath, flag="c")
        self.assertEqual(self.db["one"], "bar")
        self.assertEqual(self.db["three"], "bar")

    def test_warns_but_recovers_from_bad_checksum(self):
        self.db["one"] = "bar"
        self.db["two"] = "bar"
        self.db["three"] = "bar"
        self.db.close()

        before = os.path.getsize(self.db.filepath)

        # # Write garbage to the beginning of the second value data.
        # The first and third key should still be read and the db formed from that.
        with open(self.db.filepath, "r+b") as f:
            f.seek(13 + 8 + 18 + 18 + 4 + 8 + 18 + 18)
            f.write(b"\x00" * 4)

        with self.assertWarns(BytesWarning):
            self.db = savestate.open(filename=self.filepath, flag="c")

        self.assertEqual(self.db["one"], "bar")
        self.assertRaises(KeyError, self.db.__getitem__, "two")
        self.assertEqual(self.db["three"], "bar")

        # Test that compaction removes the corrupted data.
        self.db.close(compact=True)
        after = os.path.getsize(self.db.filepath)
        self.assertLess(after, before)

        self.db = savestate.open(filename=self.filepath, flag="c")
        self.assertEqual(self.db["one"], "bar")
        self.assertEqual(self.db["three"], "bar")

    def test_warns_but_recovers_from_missing_data_at_the_end_of_the_file(self):
        # e.g. computer crashes when writing large amounts of data
        # so parts of it were not written.
        self.db["one"] = "bar"
        self.db["two"] = "bar"
        self.db["three"] = "bar"
        self.db.close()

        before = os.path.getsize(self.db.filepath)

        self.truncate_data_file(bytes_from_end=8)

        with self.assertWarns(BytesWarning):
            self.db = savestate.open(filename=self.filepath, flag="c")

        self.assertEqual(self.db["one"], "bar")
        self.assertEqual(self.db["two"], "bar")
        self.assertRaises(KeyError, self.db.__getitem__, "three")

        # Test that compaction removes the corrupted data.
        self.db.close(compact=True)
        after = os.path.getsize(self.db.filepath)
        self.assertLess(after + 8, before)

        self.db = savestate.open(filename=self.filepath, flag="c")
        self.assertEqual(self.db["one"], "bar")

    def test_warns_but_recovers_from_trying_to_read_past_the_end_of_the_file(self):

        self.db["one"] = "bar"
        self.db["two"] = "bar"
        self.db.close()

        before = os.path.getsize(self.db.filepath)

        # Set value length indicator to 255 instead of 18
        with open(self.db.filepath, "r+b") as f:
            f.seek(13 + 8 + 18 + 18 + 4 + 4)
            f.write(b"\x00\x00\x00\xff")

        with self.assertWarns(BytesWarning):
            self.db = savestate.open(filename=self.filepath, flag="c")

        self.assertEqual(self.db["one"], "bar")
        self.assertRaises(KeyError, self.db.__getitem__, "two")

        # Test that compaction removes the corrupted data.
        self.db.close(compact=True)
        after = os.path.getsize(self.db.filepath)
        self.assertLess(after + 8, before)

        self.db = savestate.open(filename=self.filepath, flag="c")
        self.assertEqual(self.db["one"], "bar")


if __name__ == '__main__':
    unittest.main()



"""Persistent storage for arbitrary python objects inspired by SemiDBM and Shelve.

Data structure and byte sizes are as follows:
⠀\n
⠀\n
Header:⠀<file_identifier: 9 bytes> <file_format_version: 2 bytes> <pickle_version: 2 bytes>\n
Data:⠀⠀⠀<keysize: 4 bytes> <valsize: 4 bytes> <key: 'keysize' bytes> <val: 'valsize' bytes> <checksum: 4 bytes>\n
⠀⠀⠀⠀⠀⠀⠀⠀<keysize: 4 bytes> <valsize: 4 bytes> <key: 'keysize' bytes> <val: 'valsize' bytes> <checksum: 4 bytes>\n
⠀⠀⠀⠀⠀⠀⠀⠀...\n
⠀\n
"""

import os
import sys
import mmap
import uuid
import ntpath
import struct
import pickle
import builtins
import warnings

from typing import Literal, Any, Mapping, Iterator, Generator
from binascii import crc32

__all__ = [
    "open",
    "SaveStateError",
    "SaveStateLoadError",
    "SaveStateChecksumError"
]


class SaveStateError(Exception):
    pass


class SaveStateLoadError(SaveStateError):
    pass


class SaveStateChecksumError(SaveStateError):
    pass


# File handing flags
if sys.platform.startswith("win"):
    DATA_OPEN_FLAGS = os.O_RDWR | os.O_CREAT | os.O_APPEND | os.O_BINARY
    DATA_OPEN_FLAGS_READONLY = os.O_RDONLY | os.O_BINARY
else:
    DATA_OPEN_FLAGS = os.O_RDWR | os.O_CREAT | os.O_APPEND
    DATA_OPEN_FLAGS_READONLY = os.O_RDONLY

DELETED: int = 0
"""Signifies item has been deleated from the savestate."""

# Header Info
FILE_IDENTIFIER: bytes = b'savestate'
"""Magic identifier for this type of file."""
FILE_FORMAT_VERSION: int = 1
"""Version of the file format."""
PICKLE_PROTOCOL: int = 5
"""Pickle protocol used if necessary."""

# Struct (un)packing formats. Make sure that the length in bytes when
# a value is converted by the format string according to this table:
# https://docs.python.org/3/library/struct.html#format-characters
# ... is the same as the size marked here.
HEADER_FORMAT: str = "!9sHH"
HEADER_SIZE: int = 13
CHECKSUM_FORMAT: str = "!I"
CHECKSUM_SIZE: int = 4
KEYVAL_IND_FORMAT: str = "!II"
KEYVAL_IND_SIZE: int = 8


class _SaveStateReadOnly:
    """SaveState file in read-only mode, error if doesn't exist."""

    def __init__(self, filename: str, verify_checksums: bool = False, dbm_mode: bool = False):
        """Encapsulate a SaveState file in read-only mode.

        :param filename: Name of the savestate to open.
        :param verify_checksums: Verify that the checksum for a key and value pair is correct on every __getitem__ call
        :param dbm_mode: Operate in dbm mode. This is faster, but only allows strings for keys and values.
        :raises SaveStateError: Savestate file does not exist.
        """

        if not os.path.isfile(filename):
            raise SaveStateError(f"Not a file: {filename}")

        self._savestate_name = filename
        self._dbm_mode = dbm_mode
        self._data_flags = DATA_OPEN_FLAGS_READONLY
        self._verify_checksums = verify_checksums

        self._index: dict[bytes, tuple[int, int]] = self._load_index(self._savestate_name)
        """The in memory index. Index 'key' is the name of the stored value in bytes and index 'value' is a tuple 
        of the offset in bytes in the file to the stored value, and the size of the stored value in bytes."""

        self._data_file_descriptor: int = os.open(self._savestate_name, self._data_flags)
        self._current_offset: int = os.lseek(self._data_file_descriptor, 0, os.SEEK_END)

    def __repr__(self) -> str:
        selftype = type(self)
        flag = ""
        if selftype == _SaveStateReadOnly:
            flag = "r"
        elif selftype == _SaveStateReadWrite:
            flag = "w"
        elif selftype == _SaveStateCreate:
            flag = "c"
        elif selftype == _SaveStateNew:
            flag = "n"

        if hasattr(self, "_compact"):
            return f"savestate.open(filename={self._savestate_name}, flag={flag}, " \
                   f"verify_checksums={self._verify_checksums}, compact={self._compact}, dbm_mode={self._dbm_mode})"
        else:
            return f"savestate.open(filename={self._savestate_name}, flag={flag}, " \
                   f"verify_checksums={self._verify_checksums}, dbm_mode={self._dbm_mode})"

    def __str__(self) -> str:
        return "SaveState file with content: {" + ", ".join([f"'{key}': '{value}'" for key, value in self.items()]) + "}"

    def __getitem__(self, key: Any) -> Any:
        """Load value from the savestate.

        :raises KeyError: Key not found in the savestate.
        :raises AttributeError: Savestate closed.
        :raises pickle.PicklingError: Key is not pickleable.
        """

        # [sic] implementing this in _convert_to_bytes would be slower
        if self._dbm_mode:
            key = key.encode() if isinstance(key, str) else key
        else:
            key = self._convert_to_bytes(key)

        offset, size = self._index[key]
        os.lseek(self._data_file_descriptor, offset, os.SEEK_SET)

        if not self._verify_checksums:
            data = os.read(self._data_file_descriptor, size)
        else:
            data = os.read(self._data_file_descriptor, size + CHECKSUM_SIZE)
            data = self._verify_data_checksum(key, data)

        # [sic] implementing this in _convert_from_bytes would be slower
        if self._dbm_mode:
            return data
        else:
            return self._convert_from_bytes(data)

    def __iter__(self) -> Iterator[bytes]:
        for key in iter(self._index.keys()):
            yield self._convert_from_bytes(key)

    def __reversed__(self) -> Iterator[bytes]:
        for key in reversed(self._index.keys()):
            yield self._convert_from_bytes(key)

    def __contains__(self, key: Any) -> bool:
        if self._dbm_mode:
            return key.encode() if isinstance(key, str) else key in self._index
        else:
            return self._convert_to_bytes(key) in self._index

    def __len__(self):
        return len(self._index)

    def __del__(self):
        if not self.is_open:
            return  # already closed or init failed
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    @property
    def filepath(self) -> str:
        return os.path.abspath(self._savestate_name)

    @property
    def filename(self) -> str:
        return ntpath.basename(self._savestate_name)

    @property
    def is_open(self) -> bool:
        return hasattr(self, "_data_file_descriptor")

    def close(self):
        """Close the savestate.

        The data is synced to disk and the savestate is closed.
        Once the savestate has been closed, no further reads or writes are allowed.

        :raises AttributeError: Savestate closed.
        """

        os.close(self._data_file_descriptor)
        # Should be deleted to indicate file is closed to __del__
        delattr(self, "_data_file_descriptor")

    def keys(self) -> list[Any]:
        """Return all they keys in the savestate."""
        return [self._convert_from_bytes(key) for key in self._index]

    def values(self) -> list[Any]:
        """Return all they values in the savestate."""
        return [self[self._convert_from_bytes(key)] for key in self._index]

    def items(self) -> list[tuple[Any, Any]]:
        """Return list of key value pairs."""
        return [(self._convert_from_bytes(key), self[self._convert_from_bytes(key)]) for key in self._index]

    def get(self, key: Any, default: Any = None) -> Any:
        """Get value for key in savestate."""
        return self[key] if key in self else default

    def _iter_file_data(self, filename: str) -> Generator[tuple[bytes, int, int], None, None]:
        """Iterate over the stored data.

        Accepts a filename and iterates over the data bytes stored in it.
        Each yielded item should be a tuple of:

        - (key_name, offset, val_size)

        **key_name** is the name of the key (bytes).

        **offset** is the integer offset within the file to the value associated with the key.

        **val_size** is the size of the value in bytes.

        :raises SaveStateError: Something wrong with the file contents.
        """

        with builtins.open(filename, "rb") as f:
            self._verify_header(header=f.read(HEADER_SIZE))

            offset: int = HEADER_SIZE
            missing_bytes: int = 0

            with mmap.mmap(fileno=f.fileno(), length=0, access=mmap.ACCESS_READ) as contents:
                while True:

                    if offset >= len(contents):
                        break  # End of file, so stop reading values

                    try:
                        key_size, val_size = struct.unpack(KEYVAL_IND_FORMAT, contents[offset:offset + KEYVAL_IND_SIZE])
                        if key_size == 0:
                            warnings.warn(f"Zero key size at position {offset}/{len(contents)}. "
                                          f"Could not continue to read data.", category=BytesWarning)
                            break
                    except struct.error:
                        warnings.warn(f"Key and value size indicators could not be unpacked from file "
                                      f"at position {offset}/{len(contents)}.", category=BytesWarning)
                        break

                    missing_bytes = (offset + KEYVAL_IND_SIZE + key_size + val_size + CHECKSUM_SIZE) - len(contents)
                    if missing_bytes > 0:
                        warnings.warn(f"Some of the data is missing at the end of the file. "
                                      f"Compaction necessary.", category=BytesWarning)
                        break

                    offset += KEYVAL_IND_SIZE
                    key = contents[offset:offset + key_size]
                    offset += key_size

                    try:
                        self._verify_data_checksum(key, data_with_checksum=contents[offset:offset + val_size + CHECKSUM_SIZE])
                        yield key, offset, val_size

                    except SaveStateChecksumError:
                        warnings.warn(f"Data was corrupted at position {offset}/{len(contents)}. "
                                      f"Compaction necessary.", category=BytesWarning)

                    offset += (val_size + CHECKSUM_SIZE)

        # If file doesn't have enought bytes, fill with blank data to
        # recover from errors that would arise when writing more data to the file.
        if missing_bytes > 0:
            with builtins.open(filename, "ab+") as f:
                f.write(b"\x00" * missing_bytes)

    def _load_index(self, filename: str) -> dict[bytes, tuple[int, int]]:
        """This method is only used upon instantiation to populate the in memory index."""
        index = {}

        for key, offset, val_size in self._iter_file_data(filename):
            if val_size == DELETED:
                # Due to the append only nature of savestate, when values would be deleted from the file,
                # a new datapoint is appended to the file with the same key but the value size set to 'DELETED' instead.
                # This means that if val_size is DELETED, there must be a value with the same key already in the index,
                # but it should not be included, as it's marked deleted here.
                del index[key]
            else:
                index[key] = offset, val_size

        return index

    @staticmethod
    def _convert_to_bytes(value: Any) -> bytes:
        return pickle.dumps(value, protocol=PICKLE_PROTOCOL)

    @staticmethod
    def _convert_from_bytes(value: bytes) -> Any:
        return pickle.loads(value)

    @staticmethod
    def _verify_header(header: bytes):
        """Check that file is correct type and compatible version.

        :raises SaveStateError: File was incorrect type or incompatible version.
        """

        signature, file_version, pickling_version = struct.unpack(HEADER_FORMAT, header)

        if signature != FILE_IDENTIFIER:
            raise SaveStateLoadError("File is not a SaveState file.")
        if file_version != FILE_FORMAT_VERSION:
            raise SaveStateLoadError(f"Incompatible file version (got: v{file_version}, can handle: v{FILE_FORMAT_VERSION})")
        if pickling_version < PICKLE_PROTOCOL:
            raise SaveStateLoadError(f"Incompatible pickling protocol. (got: v{pickling_version}, requires: v{PICKLE_PROTOCOL})")

    @staticmethod
    def _verify_data_checksum(key: bytes, data_with_checksum: bytes) -> bytes:
        """Verify the data by calculating the checksum with CRC-32. Return data without the checksum.

        :raises SaveStateChecksumError: Checksum failed.
        """

        data_no_checksum = data_with_checksum[:-CHECKSUM_SIZE]
        checksum = struct.unpack(CHECKSUM_FORMAT, data_with_checksum[-CHECKSUM_SIZE:])[0]
        computed_checksum = crc32(key + data_no_checksum)

        if computed_checksum != checksum:
            raise SaveStateChecksumError(f"Corrupt data detected: invalid checksum for key {key}.")

        return data_no_checksum


class _SaveStateCreate(_SaveStateReadOnly):
    """SaveState file in read-write more, create if doesn't exist."""

    def __init__(self, filename: str, verify_checksums: bool = False, compact: bool = False, dbm_mode: bool = False):  # noqa
        """Encapsulate a SaveState file in read-write mode, creating a new savestate if none exists with given filename.

        :param filename: Name of the savestate to open.
        :param verify_checksums: Verify that the checksum for a key and value pair is correct on every __getitem__ call
        :param compact: Indicate whether or not to compact the savestate before closing it. No effect in read only mode.
        :param dbm_mode: Operate in dbm mode. This is faster, but only allows strings for keys and values.
        """

        self._savestate_name = filename
        self._dbm_mode = dbm_mode
        self._compact = compact
        self._data_flags = DATA_OPEN_FLAGS
        self._verify_checksums = verify_checksums

        self._index: dict[bytes, tuple[int, int]] = self._load_index(self._savestate_name)
        """The in memory index. Index 'key' is the name of the stored value in bytes and index 'value' is a tuple 
        of the offset in bytes in the file to the stored value, and the size of the stored value in bytes."""

        self._data_file_descriptor: int = os.open(self._savestate_name, self._data_flags)
        self._current_offset: int = os.lseek(self._data_file_descriptor, 0, os.SEEK_END)

    def __setitem__(self, key: Any, value: Any):
        """Save value in the savestate.

        :raises AttributeError: Savestate closed.
        :raises pickle.PicklingError: Key is not pickleable.
        """

        # [sic] implementing this in _convert_to_bytes would be slower
        if self._dbm_mode:
            key = key.encode() if isinstance(key, str) else key
            value = value.encode() if isinstance(value, str) else value
        else:
            key = self._convert_to_bytes(key)
            value = self._convert_to_bytes(value)

        key_size = len(key)
        val_size = len(value)

        keyval_size = struct.pack(KEYVAL_IND_FORMAT, key_size, val_size)
        keyval = key + value

        checksum = struct.pack(CHECKSUM_FORMAT, crc32(keyval))

        blob = keyval_size + keyval + checksum
        os.write(self._data_file_descriptor, blob)

        # Update the in memory index.
        self._index[key] = (self._current_offset + KEYVAL_IND_SIZE + key_size, val_size)
        self._current_offset += len(blob)

    def __delitem__(self, key: Any):
        """Write new value to the savestate marking that a certain key has been deleted and remove it from the index.
        When the savestate is loaded after this, it sees that the value is marked deleted and won't add it to the index.
        Still, if a value is added later under the same key, that value will be added to the index.

        :raises KeyError: Key not found in savestate.
        :raises AttributeError: Savestate closed.
        :raises pickle.PicklingError: Key is not pickleable.
        """

        # [sic] implementing this in _convert_to_bytes would be slower
        if self._dbm_mode:
            key = key.encode() if isinstance(key, str) else key
        else:
            key = self._convert_to_bytes(key)

        del self._index[key]

        key_size = struct.pack(KEYVAL_IND_FORMAT, len(key), DELETED)
        checksum = struct.pack(CHECKSUM_FORMAT, crc32(key))

        blob = key_size + key + checksum
        os.write(self._data_file_descriptor, blob)

        self._current_offset += len(blob)

    def close(self, compact: bool = False):
        """Close the savestate.

        The data is synced to disk and the savestate is closed.
        Once the savestate has been closed, no further reads or writes are allowed.

        :param compact: Enable compaction here, even if it was not enabled by open.
        :raises AttributeError: Savestate closed.
        """

        if self._compact or compact:
            self.compact()

        self.sync()
        super().close()

    def sync(self):
        """Sync the savestate to disk.

        This will flush any of the existing buffers and fsync the data to disk.

        You should call this method to guarantee that the data is written to disk.
        This method is also called whenever the savestate is `close()`'d.

        :raises AttributeError: Savestate closed.
        """

        # The files are opened unbuffered so we don't technically need to flush the file objects.
        os.fsync(self._data_file_descriptor)

    def compact(self):
        """Rewrite the contents of the savestate file.

        This method is needed because of the append only nature of the file format.
        Basically, compaction works by opening a new savestate, writing all the keys from this savestate to the new savestate, renaming the
        new savestate to the filename associated with this savestate, and reopening the new savestate as this savestate.

        Compaction is optional, since it's a trade-off between speed and storage space used.
        As a general rule of thumb, the more non-read updates you do, the more space you'll save when you compact.
        """

        # Copy the file and close it and the current file
        new_filename = self._savestate_name[:-len(FILE_IDENTIFIER) - 1] + f"_{uuid.uuid4()}"
        new_savestate = self.copy(new_filename)
        new_savestate.close()
        os.close(self._data_file_descriptor)

        # Rename the new file to the current file, replacing it in the process.
        self._rename(from_file=new_savestate._savestate_name, to_file=self._savestate_name)

        # Open the new file as the current file.
        self._index: dict[bytes, tuple[int, int]] = new_savestate._index
        self._data_file_descriptor: int = os.open(self._savestate_name, self._data_flags)
        self._current_offset: int = new_savestate._current_offset

    def clear(self):
        """Delete all data from the savestate."""
        for key in self.keys():
            self.pop(key)
        self._index = {}
        self.compact()

    def setdefault(self, key: Any, default: Any = None):
        """If key is in the savestate, return its value. If not, insert key with a value of default and return default."""

        if self._dbm_mode:
            contains = key.encode() if isinstance(key, str) else key in self._index
        else:
            contains = self._convert_to_bytes(key) in self._index

        if contains:
            return self[key]
        else:
            self[key] = default
            return default

    def pop(self, key: Any, default: Any = None):
        """If key is in the savestate, remove it and return its value, else return default if not None.

        :raises KeyError: Default is not given and key is not in the dictionary
        """

        try:
            value = self[key]
            del self[key]
            return value
        except KeyError as key_e:
            if default is not None:
                return default
            else:
                raise key_e

    def popitem(self) -> tuple[Any, Any]:
        """Get last inserted key value pair.

        :raises KeyError: Savestate empty.
        """
        key, value = self._index.popitem()
        self._index[key] = value
        key = self._convert_from_bytes(key)
        value = self.pop(key)
        return key, value

    def copy(self, new_filename: str):
        """Creates a copy of this savestate by writing all the keys from this savestate to the new savestate."""

        new_filename = _add_file_identifier(new_filename)
        assert new_filename != self._savestate_name, "Copy can't have the same filename as the original."

        new_savestate = self.__class__(filename=new_filename, verify_checksums=self._verify_checksums, compact=self._compact)

        for key in iter(self):  # Gives keys converted from bytes
            new_savestate[key] = self[key]

        return new_savestate

    def update(self, other: Mapping[Any, Any], **kwargs: Any):
        """Update the savestate with the the keys and value in other or with given kwargs.
        If both are present, kwargs will overwrite keys given in other."""

        for key, value in other.items():
            self[key] = value
        for key, value in kwargs.items():
            self[key] = value

    @staticmethod
    def _rename(from_file: str, to_file: str):
        """Renames the savestate file. If 'to_file' exists, the savestate file will replace it.

        :raises OSError: File can't be renamed. Possibly being used by another process.
        """

        if sys.platform.startswith("win"):
            import ctypes
            from ctypes.wintypes import LPVOID, DWORD

            LPCTSTR = ctypes.c_wchar_p
            kernel32 = ctypes.windll.kernel32
            kernel32.ReplaceFile.argtypes = [LPCTSTR, LPCTSTR, LPCTSTR, DWORD, LPVOID, LPVOID]

            rc = kernel32.ReplaceFile(LPCTSTR(to_file), LPCTSTR(from_file), None, 0, None, None)
            if rc == 0:
                raise OSError(f"Can't rename file, error: {kernel32.GetLastError()}")

        else:
            os.rename(from_file, to_file)

    @staticmethod
    def _write_headers(filename: str):
        """Write the header onto the file."""

        with builtins.open(filename, "wb") as f:
            f.write(struct.pack(HEADER_FORMAT, FILE_IDENTIFIER, FILE_FORMAT_VERSION, PICKLE_PROTOCOL))

    def _load_index(self, filename: str) -> dict[bytes, tuple[int, int]]:
        """This method is only used upon instantiation to populate the in memory index."""

        if not os.path.isfile(filename):
            self._write_headers(filename)
            return {}

        return super(_SaveStateCreate, self)._load_index(filename)


class _SaveStateReadWrite(_SaveStateCreate):
    """SaveState file in read-write mode, error if doesn't exist."""

    def __init__(self, filename: str, verify_checksums: bool = False, compact: bool = False, dbm_mode: bool = False):
        """Encapsulate a SaveState file in read-write mode.

        :param filename: Name of the savestate to open.
        :param verify_checksums: Verify that the checksum for a key and value pair is correct on every __getitem__ call
        :param compact: Indicate whether or not to compact the savestate before closing it. No effect in read only mode.
        :param dbm_mode: Operate in dbm mode. This is faster, but only allows strings for keys and values.
        :raises SaveStateError: Savestate file does not exist.
        """

        if not os.path.isfile(filename):
            raise SaveStateError(f"Not a file: {filename}")

        super(_SaveStateReadWrite, self).__init__(filename=filename, verify_checksums=verify_checksums, compact=compact, dbm_mode=dbm_mode)

    def copy(self, new_filename: str):
        # File needs to be opened in 'create' mode so that '__init__' does not raise 'SaveStateError' for the copy.
        # After copying, both of them can be closed and opened in 'read-write' mode.

        new_filename = _add_file_identifier(new_filename)
        assert new_filename != self._savestate_name, "Copy can't have the same filename as the original."

        self.close()
        same_savestate = open(filename=self._savestate_name, flag="c", verify_checksums=self._verify_checksums, compact=self._compact)
        new_savestate = same_savestate.copy(new_filename=new_filename)
        same_savestate.close()
        new_savestate.close()

        super(_SaveStateReadWrite, self).__init__(filename=self._savestate_name, verify_checksums=self._verify_checksums, compact=self._compact)
        new_savestate = open(filename=new_filename, flag="w", verify_checksums=self._verify_checksums, compact=self._compact)

        return new_savestate


class _SaveStateNew(_SaveStateCreate):
    """SaveState File will always be created, even if one exists."""

    def __init__(self, filename: str, verify_checksums: bool = False, compact: bool = False, dbm_mode: bool = False):
        """Encapsulate a SaveState file in read-write mode, creating a new savestate even if one exists for given filename.

        :param filename: Name of the savestate to open.
        :param verify_checksums: Verify that the checksum for a key and value pair is correct on every __getitem__ call
        :param compact: Indicate whether or not to compact the savestate before closing it. No effect in read only mode.
        :param dbm_mode: Operate in dbm mode. This is faster, but only allows strings for keys and values.
        """

        if os.path.isfile(filename):
            os.remove(filename)

        super(_SaveStateNew, self).__init__(filename=filename, verify_checksums=verify_checksums, compact=compact, dbm_mode=dbm_mode)


def _add_file_identifier(filename: str) -> str:
    """Adds SaveState file identifier to string."""
    filetype = f".{FILE_IDENTIFIER.decode()}"
    if filename[-len(FILE_IDENTIFIER) - 1:].lower() != filetype:
        filename += filetype
    return filename


def open(filename: str, flag: Literal["r", "w", "c", "n"] = "r", verify_checksums: bool = False, compact: bool = False, dbm_mode: bool = False):  # noqa
    """Open a Savestate file.

    :param filename: The name of the savestate to open. Will have the '.savestate' file extension added to it, if it doesn't have it.
    :param flag: Specifies how the savestate should be opened.
                 'r' = Open an existing savestate for reading only (default).
                 'w' = Open an existing savestate for reading and writing.
                 'c' = Open a savestate for reading and writing, creating it if it doesn't exist.
                 'n' = Always create a new, empty savestate, open for reading and writing.
    :param verify_checksums: Verify that the checksum for a key and value pair is correct on every __getitem__ call
    :param compact: Indicate whether or not to compact the savestate before closing it. No effect in read only mode.
    :param dbm_mode: Operate in dbm mode. This is faster, but only allows strings for keys and values.
    :raises ValueError: Flag argument incorrect.
    :raises SaveStateError: If flag is "r" or "w", and Savestate file does not exist.
    """

    filename = _add_file_identifier(filename)

    if flag == "r":
        return _SaveStateReadOnly(filename, verify_checksums=verify_checksums, dbm_mode=dbm_mode)
    elif flag == "w":
        return _SaveStateReadWrite(filename, verify_checksums=verify_checksums, compact=compact, dbm_mode=dbm_mode)
    elif flag == "c":
        return _SaveStateCreate(filename, verify_checksums=verify_checksums, compact=compact, dbm_mode=dbm_mode)
    elif flag == "n":
        return _SaveStateNew(filename, verify_checksums=verify_checksums, compact=compact, dbm_mode=dbm_mode)
    else:
        raise ValueError("Flag argument must be 'r', 'c', 'w', or 'n'")



"""Test SaveState performance and functionality. Be sure to run this from the command line for accuracy."""

import os
import sys
import time
import shutil
import random
import string
import argparse

from typing import Generator

__all__ = []


def _log_on_same_line(text):
    sys.stdout.write(text)
    sys.stdout.flush()


def _generate_random_data(length: int, ks: int, vs: int) -> Generator[tuple[bytes, bytes], None, None]:
    for i in range(length):

        # Display progress
        _log_on_same_line(f" {i + 1}/{length}\r")

        k = "".join(random.choice(string.ascii_letters) for _ in range(ks))  # .encode("utf-8")
        v = "".join(random.choice(string.ascii_letters) for _ in range(vs))  # .encode("utf-8")
        yield k, v


if __name__ == "__main__":

    import savestate
    import shelve

    print("\n Begin setup...\n")

    parser = argparse.ArgumentParser()
    parser.add_argument('-n', dest="num_keys", type=int, default=100_000)
    parser.add_argument('-ks', dest="keysize", type=int, default=16)
    parser.add_argument('-vs', dest="valuesize", type=int, default=100)
    parser.add_argument('-c', dest="compact", type=bool, default=True)
    args = parser.parse_args()

    print(f" Number of keys: {args.num_keys}")
    print(f" Keysize: {args.keysize}")
    print(f" Valuesize: {args.valuesize}")

    print("\n Generating random data...")
    random_data: dict[bytes, bytes] = {key: value for key, value in _generate_random_data(args.num_keys, args.keysize, args.valuesize)}

    print("\n\n Generating random read orders...")
    random_reads_one_percent = random.sample(list(random_data.keys()), args.num_keys // 100) * 100
    random_reads_all = random.sample(list(random_data.keys()), args.num_keys)

    print("\n Setup done! Begin testing...\n")

    testdir = "testdir"
    testfile = "testdb"

    # If you wish to compare performance with another library,
    # import it and add it to the list here. It must implement:
    # 'open', 'close', '__getitem__', '__setitem__' and '__delitem__'
    for d in [savestate, shelve]:
        print(" ------------------------------------------\n")
        print(f" Testing {d.__name__}\n")
        os.mkdir(testdir)
        db = d.open(os.path.join(testdir, testfile), "n")

        # --- Write random --------------------------------------------------------------

        start = time.time()
        for key, value in zip(random_reads_all, list(random_data.values())):
            db[key] = value
        total = time.time() - start

        print(f" Write time for {args.num_keys} keys randomly: {total:.2f}s.")
        print(f" {(args.num_keys / total):.0f} ops/sec.\n")

        # --- Write sequential ----------------------------------------------------------

        start = time.time()
        for key, value in random_data.items():
            db[key] = value
        total = time.time() - start

        print(f" Write time for {args.num_keys} keys linearly after random writes: {total:.2f}s.")
        print(f" {(args.num_keys / total):.0f} ops/sec.\n")

        # --- Read sequential -----------------------------------------------------------

        start = time.time()
        for key in list(random_data.keys()):
            _ = db[key]
        total = time.time() - start

        print(f" Read time for {args.num_keys} keys linearly: {total:.2f}s.")
        print(f" {(args.num_keys / total):.0f} ops/sec.\n")

        # --- Read random ---------------------------------------------------------------

        start = time.time()
        for key in random_reads_all:
            _ = db[key]
        total = time.time() - start

        print(f" Read time for {args.num_keys} keys randomly: {total:.2f}s.")
        print(f" {(args.num_keys / total):.0f} ops/sec.\n")

        # --- Read hot -------------------------------------------------------------------

        start = time.time()
        for key in random_reads_one_percent:
            _ = db[key]
        total = time.time() - start

        # Tests caching, not applicable to savestate
        print(f" Read time for random 1% of {args.num_keys} keys 100 times: {total:.2f}s.")
        print(f" {(len(random_reads_one_percent) / total):.0f} ops/sec.\n")

        # --- Delete sequential ----------------------------------------------------------

        if sys.platform.startswith("win") and d.__name__ == "shelve":
            print(" Shelve is so slow for deletion on windows it's not worth testing.\n")

        else:
            start = time.time()
            for key in list(random_data.keys()):
                del db[key]
            # SaveState will try to compact it's data (a bit slower but not that much).
            if args.compact and hasattr(db, "compact"):
                db.compact()
            total = time.time() - start

            print(f" Deleting time for {args.num_keys} keys linearly: {total:.2f}s.")
            print(f" {(args.num_keys / total):.0f} ops/sec.\n")

        # --------------------------------------------------------------------------------

        db.close()
        shutil.rmtree(testdir)
